"""
Helper scripts for running galfit
"""
import os
import glob

from collections import OrderedDict
import numpy as np

import astropy.io.fits as pyfits
import astropy.wcs as pywcs

import grizli
try:
    from .psf import DrizzlePSF
except:
    from grizli.galfit.psf import DrizzlePSF
    
GALFIT_IMAGES = """===============================================================================
# IMAGE and GALFIT CONTROL PARAMETERS
A) {input}           # Input data image (FITS file)
B) {output}       # Output data image block
C) {sigma}                # Sigma image name (made from data if blank or "none") 
D) {psf}   #        # Input PSF image and (optional) diffusion kernel
E) 1                   # PSF fine sampling factor relative to data 
F) {mask}                # Bad pixel mask (FITS image or ASCII coord list)
G) none                # File with parameter constraints (ASCII file) 
H) 1    {xmax}   1    {ymax}   # Image region to fit (xmin xmax ymin ymax)
I) {sh}    {sh}          # Size of the convolution box (x y)
J) 26.563              # Magnitude photometric zeropoint 
K) 0.06 0.06        # Plate scale (dx dy)    [arcsec per pixel]
O) regular             # Display type (regular, curses, both)
P) 0                   # Choose: 0=optimize, 1=model, 2=imgblock, 3=subcomps

# INITIAL FITTING PARAMETERS
#
#   For object type, the allowed functions are: 
#       nuker, sersic, expdisk, devauc, king, psf, gaussian, moffat, 
#       ferrer, powsersic, sky, and isophote. 
#  
#   Hidden parameters will only appear when they're specified:
#       C0 (diskyness/boxyness), 
#       Fn (n=integer, Azimuthal Fourier Modes),
#       R0-R10 (PA rotation, for creating spiral structures).
# 
# -----------------------------------------------------------------------------
#   par)    par value(s)    fit toggle(s)    # parameter description 
# -----------------------------------------------------------------------------
"""

class GalfitModel(object):
    def __str__(self, wcs=None):
        s = '# Object ID: {0}\n'.format(self.id)
        s += ' 0) {0}\n'.format(self.type)
        for k in self.pidx:
            s += '{0:>2}) '.format(self.pidx[k])
            if isinstance(self.pdict[k], list):
                s += ' '.join(['{0:.4f}'.format(v) for v in self.pdict[k]])
                if k != 'output':
                    s += ' '
                    s += ' '.join(['{0:d}'.format(v) for v in self.pfree[k]])
            else:
                s += ' {0:.4f}'.format(self.pdict[k])
                if k != 'output':
                    s += ' {0:d}'.format(self.pfree[k])
                    
            s += ' # {0}\n'.format(k)
        
        return s
        
class SersicModel(GalfitModel):
    def __init__(self, position=[0,0], magnitude=20., r_e=1., n=4., q=0.2, pa=0, output=0, id=0, wcs=None):
        self.type = 'sersic'
        self.id = 0
        
        if wcs is not None:
            self.ra, self.dec = np.array(wcs.all_world2pix([position[0]], 
                                                          [position[1]],0))[0]
                                                           
        else:
            self.ra = self.dec = None
                
        self.pidx = OrderedDict([('position',1),
                                 ('magnitude',3),
                                 ('r_e',4),
                                 ('n',5),
                                 ('q',9),
                                 ('pa',10),
                                 ('output','Z')])
                 
        self.pdict = OrderedDict([('position',position),
                                  ('magnitude',magnitude),
                                  ('r_e',r_e),
                                  ('n',n),
                                  ('q',q),
                                  ('pa',pa),
                                  ('output',output)]) 
        
        self.pfree = OrderedDict([('position',[1,1]),
                                 ('magnitude',1),
                                 ('r_e',1),
                                 ('n',1),
                                 ('q',1),
                                 ('pa',1)])

class PSFModel(GalfitModel):
    def __init__(self, position=[0,0], magnitude=20., output=0, id=0, wcs=None):
        self.type = 'psf'
        self.id = 0
        
        if wcs is not None:
            self.ra, self.dec = np.array(wcs.all_world2pix([position[0]], 
                                                          [position[1]],0))[0]
                                                           
        else:
            self.ra = self.dec = None
                
        self.pidx = OrderedDict([('position',1),
                                 ('magnitude',3),
                                 ('output','Z')])
                 
        self.pdict = OrderedDict([('position',position),
                                  ('magnitude',magnitude),
                                  ('output',output)]) 
        
        self.pfree = OrderedDict([('position',[1,1]),
                                 ('magnitude',1)])

class SkyModel(GalfitModel):
    def __init__(self, background=0, dx=0, dy=0, output=0, id=0, wcs=None, position=[0,0]):
        self.type = 'sky'
        self.id = 0
        
        if wcs is not None:
            self.ra, self.dec = np.array(wcs.all_world2pix([position[0]], 
                                                          [position[1]],0))[0]
                                                           
        else:
            self.ra = self.dec = None
                
        self.pidx = OrderedDict([('background',1),
                                 ('dx',2),
                                 ('dy',3),
                                 ('output','Z')])
                 
        self.pdict = OrderedDict([('background',background),
                                  ('dx',dx),
                                  ('dy',dy),
                                  ('output',output)]) 
        
        self.pfree = OrderedDict([('background',1),
                                 ('dx',1),
                                 ('dy',1)])
                                                                 
SERSIC="""
# Object number: 1
 0) sersic                 #  object type
 1) {x0}  {y0}  1 1  #  position x, y
 3) 20.0890     1          #  Integrated magnitude      
 4) 1.      1          #  R_e (half-light radius)   [pix]
 5) 4.      1          #  Sersic index n (de Vaucouleurs n=4) 
 6) 0.0000      0          #     ----- 
 7) 0.0000      0          #     ----- 
 8) 0.0000      0          #     ----- 
 9) 0.7570      1          #  axis ratio (b/a)  
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
 Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""

DEVN4="""
# Object number: 1
0) sersic                 #  object type
1) {x0}  {y0}  1 1  #  position x, y
3) 20.0890     1          #  Integrated magnitude      
4) 1.      1          #  R_e (half-light radius)   [pix]
5) 4.      0          #  Sersic index n (de Vaucouleurs n=4) 
6) 0.0000      0          #     ----- 
7) 0.0000      0          #     ----- 
8) 0.0000      0          #     ----- 
9) 0.7570      1          #  axis ratio (b/a)  
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""
                                                                 
NUKER="""
# Object number: 1
 0) nuker                 #  object type
 1) {x0}  {y0}  1 1  #  position x, y
 3) 20.0890     1          #  Surface brightness  
 4) 10.      1          #  Rb [pix]
 5) 4.2      1          #  Nuker powerlaw alpha
 6) 1.1      1          #     Nuker beta
 7) 0.5      1          #     Nuker gamma
 8) 0.0000      0          #     ----- 
 9) 0.7570      1          #  axis ratio (b/a)  
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
 Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""

DISK="""
# Object number: 1
 0) expdisk                 #  object type
 1) {x0}  {y0}  1 1  #  position x, y
 3) 20.0890     1          #  Surface brightness  
 4) 2.      1          #  R_s [pix]
 9) 0.7570      1          #  axis ratio (b/a)  
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
C0) -0.05 0                # diskyness(-)/boxyness(+)
 Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""

def _tostr(obj):
    if isinstance(obj, list) | isinstance(obj, tuple):
        return ' '.join(['{0}'.format(o) for o in obj])
    else:
        return '{0}'.format(obj)
         
class GalfitObject(object):
    def __repr__(self):
        return self.__str__()
    
    def __getitem__(self, k):
        return self.pdict[k]     
    
    def __setitem__(self, k, value):
        self.pdict[k] = value  
            
    def __str__(self):
        line = '# Object\n 0) {0}\n'.format(self.name)
        for k in self.pdict:
            line += '{0:>2}) {1} \t {2} \t# {3}\n'.format(self.pidx[k], _tostr(self.pdict[k]), _tostr(self.pfree[k]), k)
        
        line += '{0:>2}) {1} \t# {2}\n'.format('Z', self.output, 'output option (0 = resid., 1 = Don\'t subtract)')
        return line
        
    def set(self, **keys):
        for k in keys:
            if k in self.pdict:
                self.pdict[k] = keys[k]
    
    def setfree(self, **keys):
        for k in keys:
            if k in self.pfree:
                self.pfree[k] = keys[k]


class GalfitExpdisk(GalfitObject):
    def __init__(self, output=0, fix={}, **keys): 
        import copy
        
        self.pidx = OrderedDict([('pos', 1), 
                                 ('mag', 3), 
                                 ('R_s', 4), 
                                 ('q', 9), 
                                 ('pa', 10),
                                 ('disky', 'C0')])
                                 
        self.pdef = OrderedDict([('pos', [0,0]), 
                                    ('mag', 20.), 
                                    ('R_s', 1.), 
                                    ('q', 0.25), 
                                    ('pa', 0),
                                    ('disky', -0.05)])
                                    
        self.pfree = OrderedDict([('pos', [1,1]), 
                                  ('mag', 1), 
                                  ('R_s', 1), 
                                  ('n', 1), 
                                  ('q', 1), 
                                  ('pa', 1),
                                  ('disky', 0)])
        
        self.name = 'expdisk'
        self.output = output
        
        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)
                    
class GalfitSersic(GalfitObject):
    def __init__(self, output=0, disk=False, dev=False, fix={}, **keys): 
        import copy
        
        self.pidx = OrderedDict([('pos', 1), 
                                 ('mag', 3), 
                                 ('R_e', 4), 
                                 ('n', 5), 
                                 ('q', 9), 
                                 ('pa', 10)])
                                 
        self.pdef = OrderedDict([('pos', [0,0]), 
                                    ('mag', 20.), 
                                    ('R_e', 1.), 
                                    ('n', 4), 
                                    ('q', 0.25), 
                                    ('pa', 0)])
                                    
        self.pfree = OrderedDict([('pos', [1,1]), 
                                  ('mag', 1), 
                                  ('R_e', 1), 
                                  ('n', 1), 
                                  ('q', 1), 
                                  ('pa', 1)])
        
        self.name = 'sersic'
        self.output = output
        
        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)
        
        if disk:
            self.set(n=1)
            self.setfree(n=0)
        elif dev:
            self.set(n=4)
            self.setfree(n=0)
            
class GalfitPSF(GalfitObject):
    def __init__(self, output=0, fix={}, **keys): 
        import copy
        
        self.pidx = OrderedDict([('pos',1),
                                 ('mag',3)])
                                                  
        self.pdef = OrderedDict([('pos', [0,0]),
                                  ('mag', 20.)]) 
                                            
        self.pfree = OrderedDict([('pos', [1,1]), 
                                  ('mag', 1)])
        
        self.name = 'psf'
        self.output = output
        
        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)

class GalfitSky(GalfitObject):
    def __init__(self, output=0, fix={}, **keys): 
        import copy
                
        self.pidx = OrderedDict([('bg',1),
                                 ('dx',2),
                                 ('dy',3)])
                                                  
        self.pdef = OrderedDict([('bg', 0.),
                                  ('dx', 0.),
                                  ('dy', 0.)]) 
                                            
        self.pfree = OrderedDict([('bg', 1), 
                                  ('dx', 0),
                                  ('dy', 0)])
        
        self.name = 'sky'
        self.output = output
        
        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)
                                
class Galfitter(object):
    def __init__(self, root='sdssj1723+3411', filter='f140w', galfit_exec='galfit'):
        self.root = root
        self.galfit_exec = galfit_exec
        
        self.filter = filter
        data = self.get_data(filter=filter)
        self.sci, self.wht, self.seg, self.cat, self.wcs = data
        self.flt_files = self.get_flt_files(sci_hdu=self.sci)
        self.DPSF = DrizzlePSF(flt_files=self.flt_files, info=None, driz_image=self.sci.filename())
        
    def get_data(self, filter='f140w'):
        
        sci = pyfits.open('{0}-{1}_drz_sci.fits'.format(self.root, filter))
        wht = pyfits.open('{0}-{1}_drz_wht.fits'.format(self.root, filter))
        seg = pyfits.open('{0}-{1}_seg.fits'.format(self.root, filter))
        
        cat = grizli.utils.GTable.gread('{0}-{1}.cat'.format(self.root, filter))
        
        wcs = pywcs.WCS(sci[0].header, relax=True)
        return sci, wht, seg, cat, wcs

    def get_flt_files(self, sci_hdu=None):
        flt_files = []
        h = self.sci[0].header
        for i in range(h['NDRIZIM']):
            flt_files.append(h['D{0:03d}DATA'.format(i+1)].split('[sci')[0])
        
        return flt_files
        
    def fit_object(self, id=449, radec=(None, None), size=40, components=[GalfitSersic()], recenter=True, get_mosaic=True, gaussian_guess=False):
        """
        Fit an object
        """
        if id is not None:
            rd = self.cat['X_WORLD', 'Y_WORLD'][self.cat['NUMBER'] == id][0]
            radec = tuple(rd)
        
        xy = self.wcs.all_world2pix([radec[0]], [radec[1]], 0)
        xy = np.array(xy).flatten()
        xp = np.cast[int](np.round(xy))
        
        slx, sly, wcs_slice = self.DPSF.get_driz_cutout(ra=radec[0], dec=radec[1], get_cutout=False, size=size)
        #drz_cutout = self.get_driz_cutout(ra=ra, dec=dec, get_cutout=True)
        h = self.sci[0].header
        psf = self.DPSF.get_psf(ra=radec[0], dec=radec[1], filter=self.filter.upper(), wcs_slice=wcs_slice, pixfrac=h['D001PIXF'], kernel=h['D001KERN'], get_extended=True)
        
        exptime = h['EXPTIME']
        sci = self.sci[0].data[sly, slx]#*exptime
        wht = self.wht[0].data[sly, slx]
        rms = 1/np.sqrt(wht)#*exptime
        rms = np.sqrt((rms*exptime)**2+sci*exptime*(sci > 0))/exptime
        
        rms[wht == 0] = 1e30
        
        seg = self.seg[0].data[sly, slx]
        if id is not None:
            mask = ((seg > 0) & (seg != id)) | (wht == 0)
        else:
            mask = wht == 0
        
        sh = sci.shape[0]
        path = '/tmp/'
        
        psf_file = '{0}-{1}_psf_{2:05d}.fits'.format(self.root, self.filter, id)
        
        fp = open(path+'galfit.feedme','w')
        
        fp.write(GALFIT_IMAGES.format(input=path+'gf_sci.fits', output=path+'gf_out.fits', sigma=path+'gf_rms.fits', psf=psf_file, mask=path+'gf_mask.fits', xmax=sh, ymax=sh, sh=sh))
        
        if gaussian_guess:
            fit, q, theta = fit_gauss(sci)
        
        from astropy.coordinates import Angle
        import astropy.units as u
            
        for comp in components:
            if recenter:
                comp.set(pos=[sh/2., sh/2.])
            
            if gaussian_guess:
                comp.set(q=q, pa=Angle.wrap_at(theta*u.rad, 360*u.deg).to(u.deg).value)
                
            fp.write(str(comp))
            
        fp.close()
        
        pyfits.writeto(path+'gf_sci.fits', data=sci, overwrite=True)
        pyfits.writeto(path+'gf_rms.fits', data=rms, overwrite=True)
        pyfits.writeto(path+'gf_mask.fits', data=mask*1, overwrite=True)
        #pyfits.writeto(path+'gf_psf.fits', data=psf[1].data, overwrite=True)
        
        pyfits.writeto(psf_file, data=psf[1].data, overwrite=True)
        
        for ext in ['out', 'model']:
            if os.path.exists(path+'gf_{0}.fits'.format(ext)):
                os.remove(path+'gf_{0}.fits'.format(ext))
            
        os.system('{0} {1}/galfit.feedme'.format(self.galfit_exec, path))
        
        #out = pyfits.open('/tmp/gf_out.fits')
        
        # Get in DRZ frame
        gf_file = glob.glob('galfit.[0-9]*')[-1]
        lines = open(gf_file).readlines()
        for il, line in enumerate(lines):
            if line.startswith('A)'):
                lines[il] = 'A) {0}     # Input data image (FITS file)\n'.format(self.sci.filename())
            if line.startswith('B)'):
                lines[il] = 'B) {0}-{1}_galfit_{2:05d}.fits      # Output data image block\n'.format(self.root, self.filter, id)
            if line.startswith('P) 0'):
                lines[il] = "P) 1                   # Choose: 0=optimize, 1=model, 2=imgblock, 3=subcomps\n"
            
            if line.startswith('H)'):
                out_sh = self.sci[0].data.shape
                lines[il] = "H) 1    {0}   1    {1}   # Image region to fit (xmin xmax ymin ymax)\n".format(out_sh[1], out_sh[0])
                
            if line.startswith(' 1)'):
                xy = np.cast[float](line.split()[1:3])
                lines[il] = ' 1) {0}  {1}  1 1  #  Position x, y\n'.format(xy[0]+slx.start, xy[1]+sly.start)
                
        fp = open('galfit.{0}.{1:05d}'.format(self.filter, id),'w')
        fp.writelines(lines)
        fp.close()
        
        os.system('{0} galfit.{1}.{2:05d}'.format(self.galfit_exec, self.filter, id))
        return(lines)
        
        model = pyfits.open('/tmp/gf_model.fits')
        if False:
            full_model = self.sci[0].data*0
        
        full_model += model[0].data
        pyfits.writeto('{0}-{1}_galfit.fits'.format(self.root, self.filter), data=full_model, header=self.sci[0].header, overwrite=True)
        
def fit_gauss(sci):
    from astropy.modeling import models, fitting
    
    sh = sci.shape
    gau = models.Gaussian2D(amplitude=sci.max(), x_mean=sh[0]/2, y_mean=sh[0]/2., x_stddev=None, y_stddev=None, theta=None, cov_matrix=None)
    
    lm = fitting.LevMarLSQFitter()
    
    yp, xp = np.indices(sh)
    fit = lm(gau, xp, yp, sci)
    
    q = (fit.x_stddev/fit.y_stddev)**2
    theta = fit.theta.value
    if q > 1:
        q = 1./q
        theta += np.pi/2.
    
    return fit, q, theta
    
        
    



def link_seg_products(root='j021737-051344', filter='f140w'):
    """Symlink catalog products from the automatically-generated "ir" products to filenames for a given filter.
    
    Parameters
    ----------
    root : str
        Rootname of the mosaic images.
    
    filter : str
        Filter name.
    
    """
    import os
    os.system('ln -s {0}-ir_seg.fits {0}-{1}_seg.fits'.format(root, filter))
    os.system('ln -s {0}-ir.cat {0}-{1}.cat'.format(root, filter))
    
def galfit_model_image(root='j021737-051344', ids=[738], filter='f140w', ds9=None, ds9_frame=None, comps=None, galfit_exec='galfit', seg_threshold=0.001):
    """
    Run galfit on a list of IDs and generate model-subtracted images
    
    Parameters
    ----------
    root : str
        Rootname of the mosaic images.  Will look for images with filenames:
        
        >>> sci = '{0}-{1}_drz_sci.fits'.format(root, filter)
        >>> seg = '{0}-{1}_seg.fits'.format(root, filter)
        >>> cat = '{0}-{1}.cat'.format(root, filter)
            
    ids : list
        List of IDs in the segmentation image / catalog to fit.
        
    filter : str
        Filter name.
        
    ds9 : `~pyds9.DS9`
        DS9 instance for plotting the galfit results interactively.
        
    ds9_frame : type
        DS9 frame to load the display
        
    comps : list
        List of model component names.  If none, defaults to a two-component
        model of a Sersic plus an exponential disk.  The Sersic component is 
        initialized with `n=4`:
        
        >>> from grizli.galfit.galfit import GalfitSersic, GalfitExpdisk
        >>> comps = [GalfitSersic(n=4), galfit.GalfitExpdisk()]
        
    galfit_exec : str
        Path to the galfit executable.
    
    seg_threshold : float
        Threshold in DN (i.e., e-/s) to use for making the model segmentation
        polygons.
        
    Returns
    -------
    Creates the following images:
    
    >>> # Model-subtracted image
    >>> sub = '{0}-{1}_galfit_clean.fits'.format(root, filter)

    >>> # Model image
    >>> model = pyfits.writeto('{0}-{1}_galfit.fits'.format(root, filter)

    >>> # Segmentation image where galfit models > `seg_threshold`
    >>> gf_seg = '{0}-{1}_galfit_seg.fits'.format(root, filter)

    >>> # Segmentation image with the fitted object IDs removed
    >>> orig_seg = '{0}-{1}_galfit_orig_seg.fits'.format(root, filter)
    
    """
    import astropy.io.fits as pyfits
    from grizli.galfit import galfit
    from grizli.galfit.galfit import GalfitSersic, GalfitExpdisk
    
    if False:
        #### Testing
        reload(galfit)
        filter = 'f140w'
    
        root = 'sdssj1723+3411'
        ids = [449, 371, 286, 679, 393, 576, 225]

        root = 'sdssj2340+2947'
        ids = [486, 463]#, 439]

    gf = galfit.Galfitter(root=root, filter=filter, galfit_exec=galfit_exec) #'../Galfit/galfit')
        
    #comps = [GalfitSersic(n=4)]
    if comps is None:
        comps = [GalfitSersic(n=4), galfit.GalfitExpdisk()]
        #comps = [GalfitSersic(n=4, dev=True), galfit.GalfitExpdisk()]
        #comps = [GalfitSersic(n=4, dev=True), GalfitSersic(n=4)]
    
    # Fit galfit components
    for id in ids:
        res = gf.fit_object(id=id, radec=(None, None), size=30, get_mosaic=True, components=comps, recenter=True, gaussian_guess=False)
        
        if ds9:
            im = pyfits.open('{0}-{1}_galfit_{2:05d}.fits'.format(root, filter, id))
            if ds9_frame is not None:
                ds9.frame(ds9_frame)
            
            ds9.view(gf.sci[0].data - im[0].data/gf.sci[0].header['EXPTIME'], header=gf.sci[0].header)
    
    ###############
    # Combine components into a single image and make new segmentation images
    full_model = gf.sci[0].data*0
    full_seg = gf.seg[0].data*0
    orig_seg = gf.seg[0].data*1
    
    for id in ids:
        im = pyfits.open('{0}-{1}_galfit_{2:05d}.fits'.format(root, filter, id))
        full_model += im[0].data/gf.sci[0].header['EXPTIME']
        full_seg[(im[0].data/gf.sci[0].header['EXPTIME'] > seg_threshold) & (full_seg == 0)] = id
        orig_seg[(orig_seg == id)] = 0
        
        if ds9:
            if ds9_frame is not None:
                ds9.frame(ds9_frame)
            
            ds9.view(gf.sci[0].data - full_model)
    
    # Model-subtracted image
    pyfits.writeto('{0}-{1}_galfit_clean.fits'.format(root, filter), data=gf.sci[0].data-full_model, header=gf.sci[0].header, clobber=True)
    
    # Model image
    pyfits.writeto('{0}-{1}_galfit.fits'.format(root, filter), data=full_model, header=gf.sci[0].header, clobber=True)
    
    # Segmentation image where galfit models > `seg_threshold`
    pyfits.writeto('{0}-{1}_galfit_seg.fits'.format(root, filter), data=full_seg, header=gf.seg[0].header, clobber=True)
    
    # Segmentation image with the fitted object IDs removed
    pyfits.writeto('{0}-{1}_galfit_orig_seg.fits'.format(root, filter), data=orig_seg, header=gf.seg[0].header, clobber=True)
    


"""
Pythonic utilities ported to C [Cython] for speedup.
"""
import numpy as np
cimport numpy as np
DTYPE = np.double
ctypedef np.double_t DTYPE_t
ctypedef np.uint_t UINT_t

import cython
    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def interp_c(np.ndarray[DTYPE_t, ndim=1] x, np.ndarray[DTYPE_t, ndim=1] xp, np.ndarray[DTYPE_t, ndim=1] fp, double extrapolate=0., short int assume_sorted=1):
    """
    interp_c(x, xp, fp, extrapolate=0., assume_sorted=0)
    
    Fast interpolation: [`xp`, `fp`] interpolated at `x`.
    
    Extrapolated values are set to `extrapolate`.
    
    The default `assume_sorted`=1 assumes that the `x` array is sorted and single-
    valued, providing a significant gain in speed. (xp is always assumed to be sorted)
    
    """
    cdef unsigned long i, j, N, Np
    cdef DTYPE_t x1,x2,y1,y2,out
    cdef DTYPE_t fout, xval, xmin
    
    N, Np = len(x), len(xp)
    cdef np.ndarray[DTYPE_t, ndim=1] f = np.zeros(N)

    i=0
    j=0
    ### Handle left extrapolation
    xmin = xp[0]    
    if assume_sorted == 1:
        while x[j] < xmin: 
            f[j] = extrapolate
            j+=1
        
    while j < N:
        xval = x[j]
        if assume_sorted == 0:
            if x[j] < xmin:
                f[j] = extrapolate
                j+=1
                continue
            else:
                i=0
                
        while (xp[i] < xval) & (i < Np-1): i+=1;
        
        if i == (Np-1):
            if x[j] != xp[i]:
                f[j] = extrapolate
            else:
                f[j] = fp[i]
            j+=1
            continue   
        
        #### x[i] is now greater than xval because the 
        #### expression (x[i]<xval) is false, assuming
        #### that xval < max(x).
        
        # x1 = xp[i];
        # x2 = xp[i+1];
        # y1 = fp[i];
        # y2 = fp[i+1];
        x1 = xp[i-1];
        x2 = xp[i];
        y1 = fp[i-1];
        y2 = fp[i];
        out = ((y2-y1)/(x2-x1))*(xval-x1)+y1;
        f[j] = out
        j+=1
                
    return f
        
def interp_conserve(x, xp, fp, left=0., right=0.):
    """
    Interpolate `xp`,`yp` array to the output x array, conserving flux.  
    `xp` can be irregularly spaced.
    """
    cdef np.ndarray fullx, fully, so, outy, dx
    cdef long N, i
    
    midpoint = (x[1:]-x[:-1])/2.+x[:-1]
    midpoint = np.append(midpoint, np.array([x[0],x[-1]]))
    midpoint = midpoint[np.argsort(midpoint)]
    int_midpoint = np.interp(midpoint, xp, fp, left=left, right=right)
    int_midpoint[midpoint > xp.max()] = 0.
    int_midpoint[midpoint < xp.min()] = 0.
    
    fullx = np.append(xp, midpoint)
    fully = np.append(fp, int_midpoint)
    
    so = np.argsort(fullx)
    fullx, fully = fullx[so], fully[so]
    
    outy = x*0.
    dx = midpoint[1:]-midpoint[:-1]
    for i in range(len(x)):
        bin = (fullx >= midpoint[i]) & (fullx <= midpoint[i+1])
        outy[i] = np.trapz(fully[bin], fullx[bin])/dx[i]
        
    return outy

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def integral_cumsum_c(np.ndarray[DTYPE_t, ndim=1] xp, np.ndarray[DTYPE_t, ndim=1] fp, extrapolate=0.):
    
    cdef np.ndarray[DTYPE_t, ndim=1] ycumsum
    cdef long Nxp, i
    cdef double x0, x1, old
    
    #### Implementation of cumsum * dx
    Nxp = len(xp)
    ycumsum = np.zeros(Nxp, dtype=DTYPE)
    
    x0 = xp[0]
    ycumsum[0] = fp[0]*(xp[1]-x0)
    
    old = ycumsum[0]
    for i in range(1,Nxp):
        x1 = xp[i]
        old += fp[i]*(x1-x0)
        ycumsum[i] = old
        x0 = x1
    
    return ycumsum
    
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def new_interp_conserve_c(np.ndarray[DTYPE_t, ndim=1] x, np.ndarray[DTYPE_t, ndim=1] xp, np.ndarray[DTYPE_t, ndim=1] fp, extrapolate=0.):
    """
    Interpolate `xp`,`yp` array to the output x array, conserving flux.  
    `xp` can be irregularly spaced.
    """
    cdef np.ndarray[DTYPE_t, ndim=1] ycumsum
    cdef np.ndarray[DTYPE_t, ndim=1] inty
    cdef np.ndarray[DTYPE_t, ndim=1] outy
    
    cdef long Nx, Nxp, i
    cdef double x0, x1, old, y0, y1
    
    #### Implementation of cumsum * dx
    Nxp = len(xp)
    ycumsum = np.zeros(Nxp)
    x0 = xp[0]
    ycumsum[0] = fp[0]*(xp[1]-x0)
    
    old = ycumsum[0]
    for i in range(1,Nxp):
        x1 = xp[i]
        old += fp[i]*(x1-x0)
        ycumsum[i] = old
        x0 = x1
    
    Nx = len(x)
    inty = interp_c(x, xp, ycumsum, extrapolate=extrapolate)
    outy = np.zeros(Nx)
    
    x0 = x[0]
    y0 = inty[0]
    for i in range(1,Nx):
        y1 = inty[i]
        x1 = x[i]
        outy[i] = (y1-y0)/(x1-x0)
        x0 = x1
        y0 = y1
        
    outy[0] = outy[1]
    return outy
    
    
@cython.boundscheck(False)
def interp_conserve_c(np.ndarray[DTYPE_t, ndim=1] x, np.ndarray[DTYPE_t, ndim=1] tlam, np.ndarray[DTYPE_t, ndim=1] tf, double left=0, double right=0, double integrate=0):
    """
    interp_conserve_c(x, xp, fp, left=0, right=0, integrate=0)
    
    Interpolate `xp`,`yp` array to the output x array, conserving flux.  
    `xp` can be irregularly spaced.
    """
    cdef np.ndarray[DTYPE_t, ndim=1] templmid
    cdef np.ndarray[DTYPE_t, ndim=1] tempfmid
    cdef np.ndarray[DTYPE_t, ndim=1] outy
    cdef unsigned long i,k,istart,ntlam,NTEMPL
    cdef DTYPE_t h, numsum
    
    # templmid = (x[1:]+x[:-1])/2. #2.+x[:-1]
    # templmid = np.append(templmid, np.array([x[0], x[-1]]))
    # templmid = templmid[np.argsort(templmid)]
    NTEMPL = len(x)
    ntlam = len(tlam)

    templmid = midpoint_c(x, NTEMPL)
    #tempfmid = np.interp(templmid, tlam, tf, left=left, right=right)
    tempfmid = interp_c(templmid, tlam, tf, extrapolate=0.)
    
    outy = np.zeros(NTEMPL, dtype=DTYPE)

    ###### Rebin template grid to master wavelength grid, conserving template flux
    i=0
    k=0
    while templmid[k] < tlam[0]:
        outy[k] = left
        k+=1
        
    for k in range(k, NTEMPL):
        if templmid[k] > tlam[ntlam-1]:
            break
            
        numsum=0.;

        #### Go to where tlam is greater than the first midpoint
        while (tlam[i] < templmid[k]) & (i < ntlam): i+=1;
        istart=i;

        ####### First point
        if tlam[i] < templmid[k+1]: 
            h = tlam[i]-templmid[k];
            numsum+=h*(tf[i]+tempfmid[k]);
            i+=1;

        if i==0: i+=1;

        ####### Template points between master grid points
        while (tlam[i] < templmid[k+1]) & (i < ntlam):
            h = tlam[i]-tlam[i-1];
            numsum+=h*(tf[i]+tf[i-1]);
            i+=1;

        #### If no template points between master grid points, then just use interpolated midpoints
        if i == istart:
            h = templmid[k+1]-templmid[k];
            numsum=h*(tempfmid[k+1]+tempfmid[k]);
        else:  
            ##### Last point              
            if (templmid[k+1] == tlam[i]) & (i < ntlam):
                h = tlam[i]-tlam[i-1];
                numsum+=h*(tf[i]+tf[i-1]);
            else:
                i-=1;
                h = templmid[k+1]-tlam[i];
                numsum+=h*(tempfmid[k+1]+tf[i]);

        outy[k] = numsum*0.5;#/(templmid[k+1]-templmid[k]);
        if integrate == 0.:
            outy[k] /= (templmid[k+1]-templmid[k]);
            
    return outy
    
def midpoint(x):
    mp = (x[1:]+x[:-1])/2.
    mp = np.append(mp, np.array([x[0],x[-1]]))
    mp = mp[np.argsort(mp)]
    return mp

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def midpoint_c(np.ndarray[DTYPE_t, ndim=1] x, long N):
    cdef long i
    cdef DTYPE_t xi,xi1
    # N = len(x)
    cdef np.ndarray[DTYPE_t, ndim=1] midpoint = np.zeros(N+1, dtype=DTYPE)
    midpoint[0] = x[0]
    midpoint[N] = x[N-1]
    xi1 = x[0]
    for i in range(1, N):
        xi = x[i]
        midpoint[i] = 0.5*xi+0.5*xi1
        xi1 = xi
    
    midpoint[0] = 2*x[0]-midpoint[1]
    midpoint[N] = 2*x[N-1]-midpoint[N-1]
    
    return midpoint
    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def prepare_nmf_amatrix(np.ndarray[DTYPE_t, ndim=1] variance, np.ndarray[DTYPE_t, ndim=2] templates):
    """
    prepare_nmf_amatrix(variance, templates)
    
    Generate the "A" matrix needed for the NMF fit, which is essentially 
    T.transpose() dot T.  This function is separated from the main fitting routine 
    because it does not depend on the actual measured "flux", which the user
    might want to vary independent of the variance
    
    `templates` (T) is a 2D matrix of size (NTEMPLATE, NBAND) in terms of photo-z
    fitting ofphotometric bands.
    
    `variance` is an array with size (NBAND) representing the *measured* variance in 
    each band.  
    
    --- Cythonified from eazy/getphotz.c (G. Brammer et al. 2008) ---
    
    """
    
    cdef np.ndarray[DTYPE_t, ndim=2] amatrix
    cdef unsigned int i,j,k,NTEMP,NFILT
    
    NTEMP, NFILT = np.shape(templates)
    amatrix = np.zeros((NTEMP,NTEMP))

    for i in range(NTEMP):
        for j in range(NTEMP):
            amatrix[i,j] = 0.
            for k in range(NFILT):
                amatrix[i,j]+=templates[i,k]*templates[j,k]/variance[k]
            
    return amatrix
    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def run_nmf(np.ndarray[DTYPE_t, ndim=1] flux, np.ndarray[DTYPE_t, ndim=1] variance, np.ndarray[DTYPE_t, ndim=2] templates, np.ndarray[DTYPE_t, ndim=2] amatrix, double toler=1.e-4, long MAXITER=100000, init_coeffs=1, int verbose=0):
    """
    run_nmf(flux, variance, templates, amatrix, toler=1.e-4, MAXITER=100000, verbose=False)
    
    Run the "NMF" fit to determine the non-negative coefficients of the `templates`
    matrix that best-fit the observed `flux` and `variance` arrays.
    
    `amatrix` is generated with the `prepare_nmf_amatrix` function.
    
    e.g.
    
    >>> coeffs = run_nmf(flux, variance, templates, amatrix)
    >>> flux_fit = np.dot(coeffs.reshape((1,-1)), templates).flatten()
    >>> chi2 = np.sum((flux-flux_fit)**2/variance)
    
    --- Cythonified from eazy/getphotz.c (G. Brammer et al. 2008) ---
    """
    cdef unsigned long i,j,k,itcount, NTEMP, NFILT
    cdef double tolnum,toldenom,tol
    cdef double vold,av
    cdef np.ndarray[DTYPE_t, ndim=1] bvector
    
    cdef extern from "math.h":
        double fabs(double)

    NTEMP, NFILT = np.shape(templates)
            
    #### Make Bvector
    bvector = np.zeros(NTEMP, dtype=DTYPE)
    for i in range(NTEMP):
        bvector[i] = 0.;
        for k in range(NFILT):
            bvector[i]+=flux[k]*templates[i,k]/variance[k];
    
    #### Fit coefficients
    cdef np.ndarray[DTYPE_t, ndim=1] coeffs = np.ones(NTEMP, dtype=DTYPE)*init_coeffs
    tol = 100
    
    #### Lots of negative data, force coeffs to be zero
    for i in range(NTEMP):
        if bvector[i] < 0:
            coeffs[i] = 0.
            
    itcount=0;
    while (tol>toler) & (itcount<MAXITER):
        tolnum=0.
        toldenom = 0.
        tol=0
        for i in range(NTEMP):
            vold = coeffs[i];
            av = 0.;
            for j in range(NTEMP): av+=amatrix[i,j]*coeffs[j];
            #### Update coeffs in place      
            coeffs[i]*=bvector[i]/av;
            #tolnum+=np.abs(coeffs[i]-vold);
            tolnum+=fabs(coeffs[i]-vold);
            toldenom+=vold;

        tol = tolnum/toldenom;
        
        if verbose & 2:
            print 'Iter #%d, tol=%.2e' %(itcount, tol)
        
        itcount+=1
    
    if verbose & 1:
        print 'Iter #%d, tol=%.2e' %(itcount, tol)

    return coeffs
    
#    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.embedsignature(True)
def interpolate_tempfilt(np.ndarray[DTYPE_t, ndim=3] tempfilt, np.ndarray[DTYPE_t, ndim=1] zgrid, double zi, np.ndarray[DTYPE_t, ndim=2] output):
    """
    interpolate_tempfilt(tempfilt, zgrid, zi, output)
    
    Linear interpolate an Eazy "tempfilt" grid at z=zi.  
    
    `tempfilt` is [NFILT, NTEMP, NZ] integrated flux matrix
    `zgrid` is [NZ] redshift grid
    
    Result is stored in the input variable `output`, which needs shape [NFILT, NTEMP]
    """
    cdef unsigned long NT, NF, NZ, itemp, ifilt, iz
    cdef double dz, fint, fint2
    
    cdef extern from "math.h":
        double fabs(double)
    
    NF, NT, NZ = np.shape(tempfilt)
    
    #### Output array
    #cdef np.ndarray[DTYPE_t, ndim=2] tempfilt_interp = np.zeros((NF, NT), dtype=DTYPE)
    
    for iz in range(NZ-1):
        dz = zgrid[iz+1]-zgrid[iz]
        fint = 1 - (zi-zgrid[iz])/dz
        if (fint > 0) & (fint <= 1):
            fint2 = 1 - (zgrid[iz+1]-zi)/dz
            # print iz, zgrid[iz], fint, fint2
            for ifilt in range(NF):
                for itemp in range(NT):
                    output[ifilt, itemp] = tempfilt[ifilt, itemp, iz]*fint + tempfilt[ifilt, itemp, iz+1]*fint2
            #
            break
                    
    # return output
    



"""
Generate PSF at an arbitrary position in a drizzled image using the WFC3/IR
effective PSFs.
"""
import os
from collections import OrderedDict

import numpy as np

import astropy.io.fits as pyfits
import astropy.wcs as pywcs

#from research.dash.epsf import DrizzlePSF

DEMO_LIST = ['ib6o23rtq_flt.fits', 'ib6o23rwq_flt.fits', 'ib6o23rzq_flt.fits', 'ib6o23s2q_flt.fits']
DEMO_IMAGE = 'wfc3-ersii-g01-b6o-23-119.0-f140w_drz_sci.fits'

try:
    from .. import utils, model
except:
    from grizli import utils, model
    
class DrizzlePSF(object):
    def __init__(self, flt_files=DEMO_LIST, info=None, driz_image=DEMO_IMAGE):
        """
        Object for making drizzled PSFs
        
        Parameters
        ----------
        flt_files : list
            List of FLT files that were used to create the drizzled image.
        
        driz_image : str
            Filename of the drizzled image.
            
        """
        if info is None:
            self.wcs, self.footprint = self._get_flt_wcs(flt_files)
            self.flt_files = flt_files
        else:
            self.wcs, self.footprint = info
            self.flt_files = list(self.wcs.keys())
        
        self.ePSF = utils.EffectivePSF()
        
        self.driz_image = driz_image
        self.driz_header = pyfits.getheader(driz_image)
        self.driz_wcs = pywcs.WCS(self.driz_header)
        self.driz_pscale = utils.get_wcs_pscale(self.driz_wcs)
        
    @staticmethod
    def _get_flt_wcs(flt_files):
        """
        TBD
        """
        from shapely.geometry import Polygon, Point
        
        wcs = OrderedDict()
        footprint = OrderedDict()
        
        for file in flt_files:
            flt_j = pyfits.open(file)
            wcs[file] = pywcs.WCS(flt_j['SCI',1], relax=True)
            footprint[file] = Polygon(wcs[file].calc_footprint())
        
        return wcs, footprint
    
    def get_driz_cutout(self, ra=53.06967306, dec=-27.72333015, size=15, get_cutout=False):
        xy = self.driz_wcs.all_world2pix(np.array([[ra,dec]]), 0)[0]
        xyp = np.cast[int](np.round(xy))
        N = int(np.round(size*0.128254/self.driz_pscale))
        
        slx = slice(xyp[0]-N, xyp[0]+N)
        sly = slice(xyp[1]-N, xyp[1]+N)
        
        wcs_slice = model.ImageData.get_slice_wcs(self.driz_wcs, slx, sly)
        
        wcs_slice.pscale = utils.get_wcs_pscale(wcs_slice)
        
        # outsci = np.zeros((2*N,2*N), dtype=np.float32)
        # outwht = np.zeros((2*N,2*N), dtype=np.float32)
        # outctx = np.zeros((2*N,2*N), dtype=np.int32)
        if get_cutout:
            os.system("getfits -o sub.fits {0} {1} {2} {3} {3}".format(self.driz_image, xyp[0], xyp[1], 2*N))
            hdu = pyfits.open('sub.fits')
            return slx, sly, hdu
            
        return slx, sly, wcs_slice
    
    @staticmethod
    def _get_empty_driz(wcs):
        sh = (wcs._naxis2, wcs._naxis1)
        outsci = np.zeros(sh, dtype=np.float32)
        outwht = np.zeros(sh, dtype=np.float32)
        outctx = np.zeros(sh, dtype=np.int32)
        return outsci, outwht, outctx
        
    def go(self, ra=53.06967306, dec=-27.72333015):
        import scipy.optimize
        
        self = DrizzlePSF(info=(wcs, footprint), driz_image='cosmos-full-v1.2.8-f160w_drz_sci.fits')
        
        slx, sly, wcs_slice = self.get_driz_cutout(ra=ra, dec=dec)
        xx, yy, drz_cutout = self.get_driz_cutout(ra=ra, dec=dec, get_cutout=True)
        
        psf = self.get_psf(ra=ra, dec=dec, filter='F160W', wcs_slice=wcs_slice)
        
        init = (0, 0, drz_cutout[0].data.sum())
        chi2 = self.objfun(init, self, ra, dec, wcs_slice, filter, drz_cutout)
        
        out = scipy.optimize.minimize(self.objfun, init, args=(self, ra, dec, wcs_slice, filter, drz_cutout), method='Powell', jac=None, hess=None, hessp=None, bounds=None, constraints=(), tol=1.e-3, callback=None, options=None)
        
        psf = self.get_psf(ra=ra+out.x[0]/3600., dec=dec+out.x[1]/3600., filter=filter, wcs_slice=wcs_slice, verbose=False)
        
    @staticmethod
    def objfun(params, self, ra, dec, wcs_slice, filter, drz_cutout):
        xoff, yoff = params[:2]
        psf = self.get_psf(ra=ra+xoff/3600., dec=dec+yoff/3600., filter=filter, wcs_slice=wcs_slice, verbose=False)
        chi2 = ((psf[1].data*params[2] - drz_cutout[0].data)**2).sum()
        print(params, chi2)
        return chi2
        
    def get_psf(self, ra=53.06967306, dec=-27.72333015, filter='F140W', pixfrac=0.1, kernel='point', verbose=True, wcs_slice=None, get_extended=True,
    get_weight=False):
        from drizzlepac.astrodrizzle import adrizzle
        from shapely.geometry import Polygon, Point
        
        pix = np.arange(-13,14)
        
        #wcs_slice = self.get_driz_cutout(ra=ra, dec=dec)
        outsci, outwht, outctx = self._get_empty_driz(wcs_slice)
        
        count = 1
        for file in self.flt_files:
            if self.footprint[file].contains(Point(ra, dec)):
                if verbose:
                    print(file)
                
                xy = self.wcs[file].all_world2pix(np.array([[ra,dec]]), 0)[0]
                xyp = np.cast[int](xy)#+1
                dx = xy[0]-int(xy[0])
                dy = xy[1]-int(xy[1])
                
                psf_xy = self.ePSF.get_at_position(xy[0], xy[1], filter=filter)
                yp, xp = np.meshgrid(pix-dy, pix-dx, sparse=False, indexing='ij')
                if get_extended:
                    extended_data = self.ePSF.extended_epsf[filter]
                else:
                    extended_data = None
                    
                psf = self.ePSF.eval_ePSF(psf_xy, xp, yp, extended_data=extended_data)
                
                if get_weight:
                    fltim = pyfits.open(file)
                    flt_weight = fltim[0].header['EXPTIME']
                else:
                    flt_weight = 1
                    
                N = 13
                psf_wcs = model.ImageData.get_slice_wcs(self.wcs[file], slice(xyp[0]-N, xyp[0]+N+1), slice(xyp[1]-N, xyp[1]+N+1))
                psf_wcs.pscale = utils.get_wcs_pscale(self.wcs[file])
                
                adrizzle.do_driz(psf, psf_wcs, psf*0+flt_weight, wcs_slice, 
                                 outsci, outwht, outctx, 1., 'cps', 1,
                                 wcslin_pscale=1., uniqid=1, 
                                 pixfrac=pixfrac, kernel=kernel, fillval=0, 
                                 stepsize=10, wcsmap=None)
                
                if False:
                    count += 1
                    hdu = pyfits.HDUList([pyfits.PrimaryHDU(), pyfits.ImageHDU(data=psf*100, header=utils.to_header(psf_wcs))])                 
                    ds9.set('frame {0}'.format(count+1))
                    ds9.set_pyfits(hdu)
                
        #ss = 1000000/2
        ss = 1./outsci.sum()
        hdu = pyfits.HDUList([pyfits.PrimaryHDU(), pyfits.ImageHDU(data=outsci*ss, header=utils.to_header(wcs_slice))])
        if False:
            ds9.set('frame 2')
            ds9.set_pyfits(hdu)
        
        return hdu
